
from pathlib import Path
import time
import warnings

from IPython import embed

import numpy

from matplotlib import pyplot, ticker, rc

from astropy.io import fits

from mangadap.dapfits import DAPQualityBitMask
from mangadap.config import manga
#from mangadap.proc.util import growth_lim
from mangadap.config.analysisplan import AnalysisPlan
from mangadap.util.fileio import channel_dictionary

from mangadap.scripts import scriptbase


def init_ax(fig, pos, facecolor=None, grid=False):
    ax = fig.add_axes(pos, facecolor=facecolor)
    ax.minorticks_on()
    ax.tick_params(which='major', length=6) #, direction='in')
    ax.tick_params(which='minor', length=3) #, direction='in')
    if grid:
        ax.grid(True, which='major', color='0.8', zorder=1, linestyle='-', lw=0.5)
    return ax


def is_critical(dapver, analysis_path, daptype, plt):

    ifus = [1901, 1902, 3701, 3702, 3703, 3704, 6101, 6102, 6103, 6104, 9101, 9102,
            12701, 12702, 12703, 12704, 12705]

    dapqualbm = DAPQualityBitMask()

    critical = numpy.zeros(len(ifus), dtype=bool)

    for i,ifu in enumerate(ifus):
        # TODO: Change this to plan.method_path()
        apath = analysis_path / daptype / str(plt) / str(ifu)
        # TODO: Change this to construct_maps_file.defaults_paths()
        maps_file = apath / f'manga-{plt}-{ifu}-MAPS-{daptype}.fits.gz'

        if not maps_file.exists():
            continue

        hdu = fits.open(str(maps_file))
        critical[i] = dapqualbm.flagged(hdu[0].header['dapqual'], 'CRITICAL')

    return ifus, critical


def compile_data(dapver, analysis_path, daptype, plt):

    ifus = [1901, 1902, 3701, 3702, 3703, 3704, 6101, 6102, 6103, 6104, 9101, 9102,
            12701, 12702, 12703, 12704, 12705]

    sc_pltifu = []
    sc_snrg = []
    sc_frms = []
    sc_rchi = []

    el_pltifu = []
    el_snrg = []
    el_frms = []
    el_rchi = []

    for ifu in ifus:

        # TODO: Change this to plan.method_path()
        apath = analysis_path / daptype / str(plt) / str(ifu)
        # TODO: Change this to construct_maps_file.defaults_paths()
        maps_file = apath / f'manga-{plt}-{ifu}-MAPS-{daptype}.fits.gz'

        if not maps_file.exists():
            continue

        hdu = fits.open(str(maps_file))

        # Get the stellar data
#        uniq, indx = map(lambda x: x[1:], numpy.unique(hdu['BINID'].data[1,:,:].ravel(),
#                                                       return_index=True))
        uniq, indx = numpy.unique(hdu['BINID'].data[1,:,:].ravel(), return_index=True)
        if uniq[0] == -1:
            uniq = uniq[1:]
            indx = indx[1:]

        mask = hdu['STELLAR_VEL_MASK'].data.ravel()[indx] > 0
        indx = indx[numpy.invert(mask)]

        if len(indx) > 0:
            sc_pltifu += [ '{0}-{1}'.format(plt, ifu) ]
            sc_snrg += [ hdu['BIN_SNR'].data.ravel()[indx] ]
            sc_frms += [ hdu['STELLAR_FOM'].data[1,:,:].ravel()[indx] ]
            sc_rchi += [ hdu['STELLAR_FOM'].data[2,:,:].ravel()[indx] ]

        # Get the emission-line data
        eml = channel_dictionary(hdu, 'EMLINE_GFLUX')
#        uniq, indx = map(lambda x: x[1:], numpy.unique(hdu['BINID'].data[3,:,:].ravel(),
#                                                       return_index=True))
        uniq, indx = numpy.unique(hdu['BINID'].data[3,:,:].ravel(), return_index=True)
        if uniq[0] == -1:
            uniq = uniq[1:]
            indx = indx[1:]
        mask = hdu['EMLINE_GFLUX_MASK'].data[eml['Ha-6564'],:,:].ravel()[indx] > 0
        indx = indx[numpy.invert(mask)]

        if len(indx) > 0:
            el_pltifu += [ '{0}-{1}'.format(plt, ifu) ]
            el_snrg += [ hdu['SPX_SNR'].data.ravel()[indx] ]
            el_frms += [ hdu['EMLINE_FOM'].data[1,:,:].ravel()[indx] ]
            el_rchi += [ hdu['EMLINE_FOM'].data[2,:,:].ravel()[indx] ]

    return ifus, numpy.array(sc_pltifu), sc_snrg, sc_frms, sc_rchi, numpy.array(el_pltifu), \
                el_snrg, el_frms, el_rchi


def plate_fit_qa(dapver, analysis_path, daptype, plt):

    # Make sure the output directory exists
    # TODO: HARDCODDED PATH (i.e., this won't work for non-MaNGA data)
    plate_qa_dir = analysis_path / daptype / str(plt) / 'qa'
    if not plate_qa_dir.exists():
        plate_qa_dir.mkdir(parents=True)

    ofile = plate_qa_dir / f'{plt}-fitqa.png'
#    ofile = None

    # Determine if the analysis is a CRITICAL failure
    c_ifus, critical = is_critical(dapver, analysis_path, daptype, plt)
    pltifu = [f'{plt}-{ifu}' for ifu in c_ifus]

    # Grab the data
    d_ifus, sc_pltifu, sc_snrg, sc_frms, sc_rchi, el_pltifu, el_snrg, el_frms, el_rchi \
            = compile_data(dapver, analysis_path, daptype, plt)

    # Check if there's anything to plot
    if len(sc_pltifu) == 0 & len(el_pltifu) == 0:
        warnings.warn(f'No PLATEIFU analysis complete for plate={plt}, daptype={daptype}.')
        return

    # Basic coding check
    assert numpy.array_equal(c_ifus, d_ifus), 'Coding error'

    # Make the plot
    rc('font', size=10)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    axsc = init_ax(fig, [0.08, 0.54, 0.45, 0.45])
    axec = init_ax(fig, [0.54, 0.54, 0.45, 0.45])
    axsf = init_ax(fig, [0.08, 0.08, 0.45, 0.45])
    axef = init_ax(fig, [0.54, 0.08, 0.45, 0.45])

#    chi = numpy.concatenate(sc_rchi+el_rchi)
#    chi_lim = numpy.power(10, growth_lim(numpy.ma.log10(chi).compressed(), 0.999, fac=1.5))
#
#    frms = numpy.append(sc_frms+el_frms)
#    frms_lim = numpy.power(10, growth_lim(numpy.ma.log10(frms).compressed(), 0.999, fac=1.5))
#
#    snr = numpy.append(sc_snrg+el_snrg)
#    snr_lim = numpy.power(10, growth_lim(numpy.ma.log10(snr).compressed(), 0.999, fac=1.5))

    chi_lim = [0.3, 100]
    frms_lim = [0.005, 500]
    snr_lim = [0.8, 300]

    axsc.set_xlim(snr_lim)
    axec.set_xlim(snr_lim)
    axsf.set_xlim(snr_lim)
    axef.set_xlim(snr_lim)

    axsc.set_ylim(chi_lim)
    axec.set_ylim(chi_lim)

    axsf.set_ylim(frms_lim)
    axef.set_ylim(frms_lim)

    axsc.set_xscale('log')
    axsc.set_yscale('log')
    axec.set_xscale('log')
    axec.set_yscale('log')
    axsf.set_xscale('log')
    axsf.set_yscale('log')
    axef.set_xscale('log')
    axef.set_yscale('log')

    axsc.tick_params(axis='x', which='both', direction='in')
    axec.tick_params(axis='x', which='both', direction='in')
    axec.tick_params(axis='y', which='both', direction='in')
    axef.tick_params(axis='y', which='both', direction='in')

    axsc.xaxis.set_major_formatter(ticker.NullFormatter())
    axec.xaxis.set_major_formatter(ticker.NullFormatter())
    axec.yaxis.set_major_formatter(ticker.NullFormatter())
    axef.yaxis.set_major_formatter(ticker.NullFormatter())

    # Plot the data
    plotted = numpy.zeros(len(pltifu), dtype=bool)
    for i in range(len(pltifu)):
        # Set the point type based on the critical flag and the number
        # of pltifus plotted
        s = 20 if critical[i] else 40
        c = 'C{0}'.format(i % 10)
        lw = 0 if i < 10 else 0.5

        # Find the correct data
        sc_indx = [False] if len(sc_pltifu) == 0 else numpy.array(sc_pltifu) == pltifu[i]
        el_indx = [False] if len(el_pltifu) == 0 else numpy.array(el_pltifu) == pltifu[i]

        if numpy.sum(sc_indx) == 0 and numpy.sum(el_indx) == 0:
            continue

        plotted[i] = True

        # Plot the stellar data
        if numpy.sum(sc_indx) == 1:
            j = numpy.where(sc_indx)[0][0]
            axsc.scatter(sc_snrg[j], sc_rchi[j], marker='.', edgecolor='k', color=c, s=s, lw=lw,
                         alpha=0.7, zorder=3)
            axsf.scatter(sc_snrg[j], sc_frms[j], marker='.', edgecolor='k', color=c, s=s, lw=lw,
                         alpha=0.7, zorder=3)
        if numpy.sum(el_indx) == 1:
            j = numpy.where(sc_indx)[0][0]
            axec.scatter(el_snrg[j], el_rchi[j], marker='.', edgecolor='k', color=c, s=s, lw=lw,
                         alpha=0.7, zorder=3, label=pltifu[i])
            axef.scatter(el_snrg[j], el_frms[j], marker='.', edgecolor='k', color=c, s=s, lw=lw,
                         alpha=0.7, zorder=3)
        else:
            axec.scatter([], [], marker='.', edgecolor='k', color=c, s=s, lw=lw,
                         alpha=0.7, zorder=3, label=pltifu[i])
        
    axsc.text(-0.13, 0.5, r'$\chi^2_\nu$', ha='center', va='center', transform=axsc.transAxes,
              rotation='vertical', fontsize=12)
    axsf.text(-0.13, 0.5, r'Fractional RMS', ha='center', va='center', transform=axsf.transAxes,
              rotation='vertical', fontsize=12)
    axsf.text(0.5, -0.13, r'S/N$_g$', ha='center', va='center', transform=axsf.transAxes,
              fontsize=12)
    axef.text(0.5, -0.13, r'S/N$_g$', ha='center', va='center', transform=axef.transAxes,
              fontsize=12)

    l = axec.legend(fontsize='x-small', ncol=2, loc=2)
    if numpy.any(critical[plotted]):
        indx = numpy.where(critical[plotted])[0]
        t = l.get_texts()
        for i in indx:
            t[i].set_color('r')

    if ofile is None:
        pyplot.show()
    else:
        print('Writing: {0}'.format(ofile))
        fig.canvas.print_figure(ofile, bbox_inches='tight', dpi=100)
    fig.clear()
    pyplot.close(fig)


class PlateFitQA(scriptbase.ScriptBase):

    @classmethod
    def get_parser(cls, width=None):

        parser = super().get_parser(description='Construct per-plate QA plots', width=width)

        parser.add_argument('plate', type=int, help='plate ID to process')

        parser.add_argument('--drpver', type=str, help='DRP version', default=None)
        parser.add_argument('--dapver', type=str, help='DAP version', default=None)
        parser.add_argument("--redux_path", type=str, help="main DRP output path", default=None)
        parser.add_argument("--analysis_path", type=str, help="main DAP output path", default=None)

        parser.add_argument("--plan_file", type=str, help="parameter file with the MaNGA DAP "
                            "execution plan to use instead of the default" , default=None)

#        parser.add_argument('--daptype', type=str, help='DAP processing type', default=None)
        parser.add_argument('--normal_backend', dest='bgagg', action='store_false', default=True)

        return parser

    @staticmethod
    def main(args):
        t = time.perf_counter()
        if args.bgagg:
            pyplot.switch_backend('agg')

        # Set the paths
        redux_path = manga.drp_redux_path(drpver=args.drpver) \
                            if args.redux_path is None else Path(args.redux_path).resolve()
        analysis_path = manga.dap_analysis_path(drpver=args.drpver, dapver=args.dapver) \
                            if args.analysis_path is None else Path(args.analysis_path).resolve()

        plan = manga.MaNGAAnalysisPlan.default(analysis_path=analysis_path) \
                    if args.plan_file is None \
                    else manga.MaNGAAnalysisPlan.from_toml(args.plan_file,
                                                           analysis_path=analysis_path)

        for i, key in enumerate(plan.keys()):
            plate_fit_qa(args.dapver, analysis_path, plan[key]['key'], args.plate)

        print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))




#!/usr/bin/env python3

import os
import time
import numpy

from matplotlib import pyplot, ticker, rc

from astropy.io import fits

from mangdap.config import defaults
from mangadap.proc.util import growth_lim
from mangadap.util.bitmask import BitMask

#-----------------------------------------------------------------------------
def init_ax(fig, pos):
    ax = fig.add_axes(pos) #, facecolor='0.9')
    ax.minorticks_on()
#    ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')
    ax.minorticks_on()
    ax.tick_params(which='major', length=6) #, direction='in')
    ax.tick_params(which='minor', length=3) #, direction='in')
    return ax


def radial_coverage_histogram(dapall, plot_file=None):

    # Get the range
    arcsec_rng = growth_lim(dapall['RCOV90'], 0.99, fac=1.1)
    rore = numpy.ma.divide(dapall['RCOV90'], dapall['NSA_ELPETRO_TH50_R'])
    rore[dapall['NSA_ELPETRO_TH50_R'] < 0] = numpy.ma.masked
    rore_rng = growth_lim(rore, 0.99, fac=1.1)

    # Determine the sample
    sdssMaskbits = sdss_maskbits_file()
    targ1bm = BitMask.from_par_file(sdssMaskbits, 'MANGA_TARGET1')

    ancillary = dapall['MNGTARG3'] > 0
    primaryplus = targ1bm.flagged(dapall['MNGTARG1'],
                                  flag=['PRIMARY_v1_2_0', 'COLOR_ENHANCED_v1_2_0'])
    secondary = targ1bm.flagged(dapall['MNGTARG1'], flag='SECONDARY_v1_2_0')
    other = numpy.invert(ancillary) & numpy.invert(primaryplus) & numpy.invert(secondary)

    indx = dapall['RCOV90'] < 1
    print('Num. of observations with 90% coverage <1 arcsec: {0}'.format(numpy.sum(indx)))
    print(dapall['PLATEIFU'][indx])

    # Instantiate the figure
    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    nlim = [0.8, 1000]

    # Show RCOV90 in terms of arcseconds for each IFU
    ax = init_ax(fig, [0.1, 0.58, 0.8, 0.4])
    ax.set_ylim(nlim)
    ax.set_yscale('log', nonposy='clip')
    ifu = [ 19, 37, 61, 91, 127 ]
    for i,f in enumerate(ifu):
        indx = (dapall['IFUDESIGN']/100).astype(int) == f
        ax.hist(dapall['RCOV90'][indx], bins=50, range=arcsec_rng, alpha=0.5,
                histtype='stepfilled', color='C{0}'.format(i), zorder=2)
        ax.hist(dapall['RCOV90'][indx], bins=50, range=arcsec_rng,
                histtype='step', color='C{0}'.format(i), zorder=3)
        medp = numpy.median(dapall['RCOV90'][indx])
        ax.plot([medp, medp], nlim, color='C{0}'.format(i), zorder=4, linestyle='--')
    ax.text(-0.1, 0.5, 'N', ha='center', va='center', transform=ax.transAxes, rotation='vertical')
    ax.text(0.5, -0.13, r'$R_{90}$ (arcsec)', ha='center', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.95, '19-fiber', color='C0', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.90, '37-fiber', color='C1', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.85, '61-fiber', color='C2', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.80, '91-fiber', color='C3', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.75, '127-fiber', color='C4', ha='left', va='center', transform=ax.transAxes)

    # Show RCOV90 in terms of Re for each sample
    ax = init_ax(fig, [0.1, 0.08, 0.8, 0.4])
    ax.set_ylim(nlim)
    ax.set_yscale('log', nonposy='clip')
    ax.hist(rore[primaryplus], bins=50, range=rore_rng, alpha=0.5, histtype='stepfilled',
            color='C0', zorder=2)
    ax.hist(rore[primaryplus], bins=50, range=rore_rng, histtype='step',
            color='C0', zorder=3)
    medp = numpy.median(rore[primaryplus])
    ax.plot([medp, medp], nlim, color='C0', zorder=4, linestyle='--')
    ax.hist(rore[secondary], bins=50, range=rore_rng, alpha=0.5, histtype='stepfilled',
            color='C1', zorder=2)
    ax.hist(rore[secondary], bins=50, range=rore_rng, histtype='step',
            color='C1', zorder=3)
    medp = numpy.median(rore[secondary])
    ax.plot([medp, medp], nlim, color='C1', zorder=4, linestyle='--')
    ax.hist(rore[other], bins=50, range=rore_rng, alpha=0.5, histtype='stepfilled',
            color='C2', zorder=2)
    ax.hist(rore[other], bins=50, range=rore_rng, histtype='step',
            color='C2', zorder=3)
    ax.hist(rore[ancillary], bins=50, range=rore_rng, alpha=0.5, histtype='stepfilled',
            color='C3', zorder=2)
    ax.hist(rore[ancillary], bins=50, range=rore_rng, histtype='step',
            color='C3', zorder=3)
    ax.text(-0.1, 0.5, 'N', ha='center', va='center', transform=ax.transAxes, rotation='vertical')
    ax.text(0.5, -0.12, r'$R_{90}/R_{eff}$', ha='center', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.95, 'Primary+', color='C0', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.90, 'Secondary', color='C1', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.85, 'Other', color='C2', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.80, 'Ancillary', color='C3', ha='left', va='center', transform=ax.transAxes)

    if plot_file is None:
        pyplot.show()
    else:
        fig.canvas.print_figure(plot_file, bbox_inches='tight')
    fig.clear()
    pyplot.close(fig)


#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.clock()

    dapall_file = '/Users/westfall/Work/MaNGA/testing/dr15/dapall-v2_4_3-2.2.1.fits'
    dapall_hdu = fits.open(dapall_file)

    radial_coverage_histogram(dapall_hdu['DAPALL'].data,
                              plot_file='../ms/figs/radial_coverage.pdf')

    print('Elapsed time: {0} seconds'.format(time.clock() - t))






import os
import numpy as np
import astropy.constants
from astropy.io import fits
import matplotlib.pyplot as pl
import matplotlib.cm as cm
import matplotlib.colors as mcol

from mangadap.dapfits import DAPCubeBitMask
from mangadap.util.fileio import channel_dictionary
import marvin.utils.plot.map as mapplot
import pdb


plate = 7443
ifu = 1902
nsa_redshift = 0.0189259
vel = nsa_redshift * astropy.constants.c.to('km/s').value
plate_str = "{:.0f}".format(plate)
ifu_str = "{:.0f}".format(ifu)
outfil = 'check_new_binning_scheme_'+plate_str+'-'+ifu_str+'.pdf'

analysis_path = os.getcwd()+'/5x5n-GAU-MILESHC/'
fits_cube_fil = analysis_path+plate_str+'/'+ifu_str+'/manga-'+plate_str+'-'+ifu_str+'-LOGCUBE-5x5n-GAU-MILESHC.fits.gz'
fits_maps_fil = analysis_path+plate_str+'/'+ifu_str+'/manga-'+plate_str+'-'+ifu_str+'-MAPS-5x5n-GAU-MILESHC.fits.gz'
#pdb.set_trace()

# Read in CUBE
hdu_cube = fits.open(fits_cube_fil)
bm = DAPCubeBitMask()

wave = hdu_cube['WAVE'].data
flux = np.ma.MaskedArray(hdu_cube['FLUX'].data,
                         mask=bm.flagged(hdu_cube['MASK'].data,
                        ['IGNORED', 'FLUXINVALID', 'IVARINVALID', 'ARTIFACT']))
ivar = np.ma.MaskedArray(hdu_cube['IVAR'].data,
                         mask=bm.flagged(hdu_cube['MASK'].data,
                        ['IGNORED', 'FLUXINVALID', 'IVARINVALID', 'ARTIFACT']))
model = np.ma.MaskedArray(hdu_cube['MODEL'].data,
                          mask=bm.flagged(hdu_cube['MASK'].data, 'FITIGNORED'))

# Read in MAP
hdu_maps = fits.open(fits_maps_fil)
binid = hdu_maps['BINID'].data[1,:,:]
mask_ext = hdu_maps['EMLINE_GVEL'].header['QUALDATA']
emlc = channel_dictionary(hdu_maps, 'EMLINE_GVEL')
emission_vfield = np.ma.MaskedArray(hdu_maps['EMLINE_GVEL'].data[emlc['Ha-6564'],:,:],
                                    mask=hdu_maps[mask_ext].data[emlc['Ha-6564'],:,:] > 0)
emission_vfield_ivar = np.ma.MaskedArray(hdu_maps['EMLINE_GVEL_IVAR'].data[emlc['Ha-6564'],:,:],
                                        mask=hdu_maps[mask_ext].data[emlc['Ha-6564'],:,:] > 0)


mask_ext = hdu_maps['STELLAR_VEL'].header['QUALDATA']
stellar_vfield = np.ma.MaskedArray(hdu_maps['STELLAR_VEL'].data,
                                    mask=hdu_maps[mask_ext].data > 0)
stellar_vfield_ivar = np.ma.MaskedArray(hdu_maps['STELLAR_VEL_IVAR'].data,
                                        mask=hdu_maps[mask_ext].data > 0)


#Plot map of binids, emission line velocities, etc
fig, axes = pl.subplots(1, 3, figsize=(12,4))

rand_cmap = mcol.ListedColormap(np.random.rand(400,3))
mapplot.plot(value=binid, title='Bin ID', cmap=rand_cmap, fig=fig, ax=axes[0], cbrange=[0,315])

mapplot.plot(value=emission_vfield, title='Emission Line Velocity', cmap=cm.coolwarm,
             cbrange=[-120,120], fig=fig, ax=axes[1])

mapplot.plot(value=stellar_vfield, title='Stellar Velocity', cmap=cm.coolwarm,
             cbrange=[-120,120], fig=fig, ax=axes[2])

fig.tight_layout()
pl.savefig(outfil)
pdb.set_trace()


from IPython import embed

import numpy
from scipy import special

import astropy.constants

from mangadap.proc.spectralstack import SpectralStack


def pixelated_gaussian(x, c=0.0, s=1.0):
    """
    x must be linearly sampled

    x, c, and s must be in the same units.
    """
    n = numpy.sqrt(2.)*s
    d = numpy.asarray(x)-c
    dx = numpy.mean(numpy.diff(x))
    return (special.erf((d+dx/2.)/n) - special.erf((d-dx/2.)/n))/2.


def test_register():

    # Generate some line centers and redshifts
    nlines = 10
    nspec = 20
    rng = numpy.random.default_rng()
    line_flux = rng.uniform(low=10, high=100, size=nlines)
    center = rng.uniform(low=3650., high=10000, size=nlines)
    z = rng.uniform(low=0.03, high=0.1, size=20)
    cz = z*astropy.constants.c.to('km/s').value
    sigma = 1 # In pixels

    # Make the wavelength vector
    wave = numpy.logspace(*numpy.log10([3600., 10300.]).tolist(), 3000)
    # Some convenience things to make sure the Gaussian profiles are sampled in
    # pixels
    dlogw = numpy.mean(numpy.diff(numpy.log10(wave)))
    coff = numpy.log10(wave[0])/dlogw
    x = numpy.arange(wave.size)

    # Construct the spectra
    flux = numpy.zeros((nspec,wave.size), dtype=float)
    for i in range(nspec):
        _center = center*(1+z[i])
        _center = numpy.log10(_center)/dlogw - coff
        for f,c in zip(line_flux,_center):
            flux[i] += f * pixelated_gaussian(x, c=c, s=sigma)

    # Stacking object
    stacker = SpectralStack()
    # Stack the spectra including the offset by the input velocity
    swave, sflux, sfdev, snpix, sivar, ssres, scovar = stacker.stack(wave, flux, cz=cz, log=True)

    # Just register the spectra to make sure the registration works;
    # this is a simple check on the above
    rwave, rflux, rivar, rsres = SpectralStack.register(wave, cz, flux, log=True)
    reg_stack_flux = numpy.ma.mean(rflux, axis=0)

    assert numpy.allclose(sflux[0], reg_stack_flux), 'Stack and by-hand check failed.'

    # Construct the deredshifted stack from scratch
    rcoff = numpy.log10(rwave[0])/dlogw
    rx = numpy.arange(rwave.size)
    model_flux = numpy.zeros(reg_stack_flux.size, dtype=float)
    rcenter = numpy.log10(center)/dlogw - rcoff
    for f,c in zip(line_flux,rcenter):
        model_flux += f * pixelated_gaussian(rx, c=c, s=sigma)

    # Compare the stacks against truth
#    from matplotlib import pyplot
#    pyplot.plot(swave, sflux[0])            # Stacked using stacker.stack (blue)
#    pyplot.plot(rwave, reg_stack_flux)      # Stacked by hand after registration (orange)
#    pyplot.plot(rwave, model_flux)          # Truth (green)
#    pyplot.show()

    xcor = numpy.correlate(sflux[0], model_flux)
    assert numpy.argmax(xcor) == xcor.size//2, 'Should be no lag between the model and stack'
    assert numpy.absolute(numpy.sum(sflux[0])/numpy.sum(model_flux) - 1) < 0.01, \
            'Stack sums should be different by less than 1%'





import pytest
import os
import subprocess
import shutil

from IPython import embed

from mangadap.tests.util import remote_data_file, data_test_file, drp_test_version
from mangadap.tests.util import requires_remote, requires_drpcomplete, requires_drpall

from mangadap.scripts.calculate_covariance import CalculateCovariance
from mangadap.scripts.manga_dap import MangaDap
from mangadap.scripts.write_dap_config import WriteDapConfig
from mangadap.util.parser import DefaultConfig


def script_help_okay(executable):
    # Basic help call
    return subprocess.call([executable, '-h'], stdout=subprocess.DEVNULL,
                           stderr=subprocess.DEVNULL) == 0

@requires_remote
def test_calculate_covariance():

    assert script_help_okay('dap_calculate_covariance'), 'Basic help call failed'

    ofile = 'test.fits'
    if os.path.isfile(ofile):
        # Clean-up a previous failure
        os.remove(ofile)

    # Defaults to central channel
    # TODO: Test contents?
    CalculateCovariance.main(CalculateCovariance.parse_args(['7815', '3702', ofile, '-d',
                                                             remote_data_file()]))
    assert os.path.isfile(ofile), 'Output file not written.'

    # Do a number of channels
    os.remove(ofile)
    CalculateCovariance.main(CalculateCovariance.parse_args(['7815', '3702', ofile, '-n', '11',
                                                             '-d', remote_data_file()]))
    assert os.path.isfile(ofile), 'Output file not written.'

    # Run a specific wavelength
    os.remove(ofile)
    CalculateCovariance.main(CalculateCovariance.parse_args(['7815', '3702', ofile, '-w',
                                                             '4500.', '-d', remote_data_file()]))
    assert os.path.isfile(ofile), 'Output file not written.'

    # Clean-up
    os.remove(ofile)


# TODO: Add some remote files to test this?
def test_construct_dapall():
    assert script_help_okay('construct_dapall'), 'Basic help call failed'


# TODO: Include this in the construct_dapall test and test the
# constructed file?
def test_dapall_qa():
    assert script_help_okay('dapall_qa'), 'Basic help call failed'


# TODO: Include a DRPComplete file in remote?
def test_find_repeat_observations():
    assert script_help_okay('dap_find_repeat_observations'), 'Basic help call failed'


# TODO: Add some remote files?
def test_fit_residuals():
    assert script_help_okay('dap_fit_residuals'), 'Basic help call failed'


# TODO: Add some remote files?
def test_inspector():
    assert script_help_okay('manga_dap_inspector'), 'Basic help call failed'


#def test_manga_dap_import():
#    # TODO: I don't understand why these exceptions changed!
#    with pytest.raises(SystemExit):
#        MangaDap.main(MangaDap.parse_args(['-c', data_test_file('datacube.ini'), '-m', 'junk']))
#    with pytest.raises(SystemExit):
#        MangaDap.main(MangaDap.parse_args(['-c', data_test_file('datacube.ini'), '-o', 'junk']))

#    with pytest.raises(ImportError):
#        MangaDap.main(MangaDap.parse_args(['-c', data_test_file('datacube.ini'), '-m', 'junk']))
#    with pytest.raises(AttributeError):
#        MangaDap.main(MangaDap.parse_args(['-c', data_test_file('datacube.ini'), '-o', 'junk']))


@requires_remote
def test_manga_dap():
    odir = data_test_file('manga_dap_output')

    # Clean up previous failure
    if os.path.isdir(odir):
        shutil.rmtree(odir)

    # Run the DAP. The binning in plan.par is set to ALL binning, so
    # this run of the DAP just analyzes one spectrum and takes about a
    # minute.
    MangaDap.main(MangaDap.parse_args(['-c', data_test_file('datacube.ini'),
                                       '-p', data_test_file('global_bin.toml'), '-o', odir]))
                                       #'-p', data_test_file('plan.par'), '-o', odir]))

    # Re-run to use existing files.  Takes about 40s.
    MangaDap.main(MangaDap.parse_args(['-c', data_test_file('datacube.ini'),
                                       '-p', data_test_file('global_bin.toml'), '-o', odir]))
                                       #'-p', data_test_file('plan.par'), '-o', odir]))

    # Clean up
    shutil.rmtree(odir)


# TODO: Add some remote data?
def test_plate_fit_qa():
    assert script_help_okay('dap_plate_fit_qa'), 'Basic help call failed'


# TODO: Add some remote data?
def test_ppxffit_qa():
    assert script_help_okay('dap_ppxffit_qa'), 'Basic help call failed'


# TODO: Add some remote data?
def test_spotcheck():
    assert script_help_okay('spotcheck_dap_maps'), 'Basic help call failed'


# TODO: Add some remote data?
def test_template_flux_norm():
    assert script_help_okay('dap_template_flux_norm'), 'Basic help call failed'


# TODO: Add a remote DRPComplete file!
def test_write_dap_config_basic():
    assert script_help_okay('write_dap_config'), 'Basic help call failed'


@requires_drpcomplete
def test_write_dap_config_drpcomplete():
    ofile = 'test.ini'
    if os.path.isfile(ofile):
        # Clean-up a previous failure
        os.remove(ofile)

    drpc = remote_data_file('drpcomplete_{0}.fits'.format(drp_test_version))
    WriteDapConfig.main(WriteDapConfig.parse_args(['7815', '3702', ofile, '-c', drpc]))

    assert os.path.isfile(ofile), 'Output file not written.'

    cfg = DefaultConfig(ofile)
    assert cfg.getint('plate') == 7815, 'Wrong plate number'
    assert cfg.get('directory_path') is None, 'No directory path should be defined.'
    assert cfg.getfloat('z') == 0.0293823, 'Bad redshift'

    # Clean-up a previous failure
    os.remove(ofile)


@requires_drpall
def test_write_dap_config_drpall():
    ofile = 'test.ini'
    if os.path.isfile(ofile):
        # Clean-up a previous failure
        os.remove(ofile)

    drpall = remote_data_file('drpall-{0}.fits'.format(drp_test_version))
    WriteDapConfig.main(WriteDapConfig.parse_args(['7815', '3702', ofile, '-a', drpall]))

    assert os.path.isfile(ofile), 'Output file not written.'

    cfg = DefaultConfig(ofile)
    assert cfg.getint('plate') == 7815, 'Wrong plate number'
    assert cfg.get('directory_path') is None, 'No directory path should be defined.'
    assert cfg.getfloat('z') == 0.0293823, 'Bad redshift'

    # Clean-up a previous failure
    os.remove(ofile)


# TODO: Add some remote files?
def test_status():
    assert script_help_okay('dap_status'), 'Basic help call failed'


# TODO: Improve this!
def test_run():
    assert script_help_okay('rundap'), 'Basic help call failed'




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""

Provides a set of utility functions dealing with computational geometry.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

import numpy
from scipy import linalg

def polygon_winding_number(polygon, point):
    """
    Determine the winding number of a 2D polygon about a point.  The
    code does **not** check if the polygon is simple (no interesecting
    line segments).  Algorithm taken from Numerical Recipies Section
    21.4.

    Args:
        polygon (numpy.ndarray): An Nx2 array containing the x,y
            coordinates of a polygon.  The points should be ordered
            either counter-clockwise or clockwise.
        point (numpy.ndarray): A 2-element array defining the x,y
            position of the point to use as a reference for the winding
            number.

    Returns:
        int: Winding number of `polygon` w.r.t. `point`

    Raises:
        ValueError: Raised if `polygon` is not 2D, if `polygon` does not
            have two columns, or if `point` is not a 2-element array.
    """

    # Check input shape is for 2D only
    if len(polygon.shape) != 2:
        raise ValueError('Polygon must be an Nx2 array.')
    if polygon.shape[1] != 2:
        raise ValueError('Polygon must be in two dimensions.')
    if point.size != 2:
        raise ValueError('Point must contain two elements.')

    # Get the winding number
    np=polygon.shape[0]
    x0 = polygon[np-1,0]
    y0 = polygon[np-1,1]
    wind = 0
    for i in range(np):
        if (y0 > point[1]):
            if polygon[i,1] <= point[1] and \
               (x0-point[0])*(polygon[i,1]-point[1]) - (y0-point[1])*(polygon[i,0]-point[0]) < 0:
                wind -= 1
        else:
            if polygon[i,1] > point[1] and \
               (x0-point[0])*(polygon[i,1]-point[1]) - (y0-point[1])*(polygon[i,0]-point[0]) > 0:
                wind += 1
        x0 = polygon[i,0]
        y0 = polygon[i,1]

    return wind


def point_inside_polygon(polygon, point):
    """
    Determine if a point is inside a polygon using the winding number.

    Args:
        polygon (numpy.ndarray): An Nx2 array containing the x,y
            coordinates of a polygon.  The points should be ordered
            either counter-clockwise or clockwise.
        point (numpy.ndarray): A 2-element array defining the x,y
            position of the point to use as a reference for the winding
            number.

    Returns:
        bool: True if the point is inside the polygon.

    .. warning:: 
        If the point is **on** the polygon (or very close to it w.r.t.
        the machine precision), the returned value is `False`.

    """
    _point = numpy.atleast_2d(point)
    if _point.shape[-1] != 2:
        raise ValueError('Provided point must have two elements in last dimension.')
    if _point.shape[0] == 1:
        return (abs(polygon_winding_number(polygon, point)) == 1)
    return numpy.array([ abs(polygon_winding_number(polygon, p)) == 1 for p in _point ])


def polygon_area(x, y):
    """
    Compute the area of a polygon using the `Shoelace formula
    <https://en.wikipedia.org/wiki/Shoelace_formula>`_.

    Inspired by `this discussion
    <https://stackoverflow.com/questions/24467972/calculate-area-of-polygon-given-x-y-coordinates>`_.

    Args:
        x (`numpy.ndarray`_):
            Vector with the Cartesian x-coordinates of the polygon
            vertices.
        y (`numpy.ndarray`_):
            Vector with the Cartesian y-coordinates of the polygon
            vertices.
            
    Returns:
        :obj:`float`: Polygon area
    """
    return 0.5 * numpy.abs(numpy.dot(x, numpy.roll(y, 1)) - numpy.dot(y, numpy.roll(x, 1)))


class SemiMajorAxisCoo:
    r"""
    Calculate the semi-major axis coordinates given a set of input
    parameters following :math:`{\mathbf x} = {\mathbf A}^{-1}\ {\mathbf
    b}`, where

    .. math::

        {\mathbf A} = \left[
        \begin{array}{rrrrrr}
            1 & 0 & 0 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 & 0 & 0 \\
            \cos\psi & \sin\psi & -1 & 0 & 0 & 0 \\
            -\sin\psi & \cos\psi & 0 & -1 & 0 &  0 \\
            0 & 0 & \sin\phi_0 & \cos\phi_0 & -1 & 0 \\
            0 & 0 & -\cos\phi_0 & \sin\phi_0 & 0 & \varepsilon-1
        \end{array}
        \right]

        {\mathbf b} = \left[
        \begin{array}{r}
            x_f \\
            y_f \\
            -x_0 \\
            -y_0 \\
            0 \\
            0 
        \end{array}
        \right]

    such that

    .. math::

        {\mathbf x} = \left[
        \begin{array}{r}
            x_f \\
            y_f \\
            x_s \\
            y_s \\
            x_a \\
            y_a 
        \end{array}
        \right]

    and:
        - :math:`\psi` is the Cartesian rotation of the focal-plane
          relative to the sky-plane (+x toward East; +y toward North),
        - :math:`\phi_0` is the on-sky position angle of the major axis
          of the ellipse, defined as the angle from North through East
        - :math:`\varepsilon=1-b/a` is the ellipticity based on the the
          semi-minor to semi-major axis ratio (:math:`b/a`).
        - :math:`(x_f,y_f)` is the sky-right, focal-plane position
          relative to a reference on-sky position :math:`(x_0,y_0)`
          relative to the center of the ellipse (galaxy center),
        - :math:`(x_s,y_s)` is the on-sky position of :math:`(x_f,y_f)`
          relative to the center of the ellipse, and
        - :math:`(x_a,y_a)` is the Cartesian position of
          :math:`(x_f,y_f)` in units of the semi-major axis.

    This form is used such that :math:`{\mathbf A}` need only be defined
    once per class instance.

    The class also allows for inverse calculations, i.e., calculating
    the focal-plane positions provide the semi-major axis coordinates.
    In this case,

    .. math::

        {\mathbf C} = \left[
        \begin{array}{rrrr}
            \cos\psi & \sin\psi & -1 & 0 \\
            -\sin\psi & \cos\psi & 0 & -1  \\
            0 & 0 & \sin\phi_0 & \cos\phi_0 \\
            0 & 0 & -\cos\phi_0 & \sin\phi_0
        \end{array}
        \right]

        {\mathbf d} = \left[
        \begin{array}{r}
            -x_0 \\
            -y_0 \\
            x_a \\
            y_a (1-\varepsilon)
        \end{array}
        \right]

    such that

    .. math::

        {\mathbf f} = \left[
        \begin{array}{r}
            x_f \\
            y_f \\
            x_s \\
            y_s
        \end{array}
        \right]


    and :math:`{\mathbf f} = {\mathbf C}^{-1}\ {\mathbf d}`.

    Args:
        xc (float): Same as :math:`x_0`, defined above
        yc (float): Same as :math:`y_0`, defined above
        rot (float): Same as :math:`\psi`, defined above
        pa (float): Same as :math:`\phi_0`, defined above
        ell (float): Same as :math:`\varepsilon`, defined above

    Attributes:
        xc,yc (float,float): a reference on-sky position relative to the
            center of the ellipse (galaxy center); same as
            :math:`(x_0,y_0)` defined above
        rot (float): Cartesian rotation of the focal-plane relative to
            the sky-plane (+x toward East; +y toward North); same as
            :math:`\psi` defined above
        pa (float): On-sky position angle of the major axis of the
            ellipse, defined as the angle from North through East and is
            the same as :math:`\phi_0` defined above
        ell (float):  Ellipticity define as :math:`\varepsilon=1-b/a`,
            based on the semi-minor to semi-major axis ratio
            (:math:`b/a`) of the ellipse.
        A (numpy.ndarray): The coordinate transformation matrix
        Alu (numpy.ndarray): The **lu** array returned by
            `scipy.linalg.lu_factor`_, which is used to calculate the LU
            decomposition of :math:`{\mathbf A}` 
        Apiv (numpy.ndarray): The **piv** array returned by
            `scipy.linalg.lu_factor`_, which is used to calculate the LU
            decomposition of :math:`{\mathbf A}` 
        B (numpy.ndarray): The vector :math:`{\mathbf b}`, as defined
            above, used to calculate :math:`{\mathbf x} = {\mathbf
            A}^{-1}\ {\mathbf b}`
        C (numpy.ndarray): The coordinate transformation matrix use for
            the inverse operations
        Clu (numpy.ndarray): The **lu** array returned by
            `scipy.linalg.lu_factor`_, which is used to calculate the LU
            decomposition of :math:`{\mathbf C}` 
        Cpiv (numpy.ndarray): The **piv** array returned by
            `scipy.linalg.lu_factor`_, which is used to calculate the LU
            decomposition of :math:`{\mathbf C}` 
        D (numpy.ndarray): The vector :math:`{\mathbf d}`, as defined
            above, used to calculate :math:`{\mathbf f} = {\mathbf
            C}^{-1}\ {\mathbf d}`
        
    """
    def __init__(self, xc=None, yc=None, rot=None, pa=None, ell=None):

        self.xc = 0.0 if xc is None else xc
        self.yc = 0.0 if yc is None else yc
        self.rot = 0.0 if rot is None else rot
        self.pa = 0.0 if pa is None else pa
        self.ell = 0.0 if ell is None else ell

        self.A = None
        self.Alu = None
        self.Apiv = None
        self.B = None

        self._setA()

        self.C = None
        self.Clu = None
        self.Cpiv = None
        self.D = None

        self._setC()


    def _defined(self):
        """
        Determine if the object is defined such that its methods can be
        used to convert between coordinate systems.
        """
        if self.A is None:
            return False
        if self.Alu is None:
            return False
        if self.Apiv is None:
            return False
        if self.C is None:
            return False
        if self.Clu is None:
            return False
        if self.Cpiv is None:
            return False
        return True
    

    def _setA(self):
        """
        Set the transformation matrix and calculate its LU
        decomposition for forward operations.
        """
        cosr = numpy.cos( numpy.radians(self.rot) )
        sinr = numpy.sin( numpy.radians(self.rot) )
        cosp = numpy.cos( numpy.radians(self.pa) )
        sinp = numpy.sin( numpy.radians(self.pa) )
        #cosi = numpy.cos( numpy.radians(self.inc) )

        self.A = numpy.array([ [   1.0,  0.0,   0.0,  0.0,  0.0,   0.0 ],
                               [   0.0,  1.0,   0.0,  0.0,  0.0,   0.0 ],
                               [  cosr, sinr,  -1.0,  0.0,  0.0,   0.0 ],
                               [ -sinr, cosr,   0.0, -1.0,  0.0,   0.0 ],
                               [   0.0,  0.0,  sinp, cosp, -1.0,   0.0 ],
                               [   0.0,  0.0, -cosp, sinp,  0.0, self.ell-1. ] ])

        self.Alu, self.Apiv = linalg.lu_factor(self.A)
        

    def _setB(self, x, y):
        """
        Set the on-sky coordinate vector for forward operations.

        Args:
            x, y (float) : Single values for use in calculating the
                semi-major-axis coordinates.
        """
        self.B = numpy.array([ x, y, -self.xc, -self.yc, 0.0, 0.0 ])


    def _setC(self):
        """
        Set the transformation matrix and calculate its LU
        decomposition for inverse operations.
        """
        cosr = numpy.cos( numpy.radians(self.rot) )
        sinr = numpy.sin( numpy.radians(self.rot) )
        cosp = numpy.cos( numpy.radians(self.pa) )
        sinp = numpy.sin( numpy.radians(self.pa) )

        self.C = numpy.array([ [  cosr, sinr,  -1.0,  0.0 ],
                               [ -sinr, cosr,   0.0, -1.0 ],
                               [   0.0,  0.0,  sinp, cosp ],
                               [   0.0,  0.0, -cosp, sinp ] ])

        self.Clu, self.Cpiv = linalg.lu_factor(self.C)
        

    def _setD(self, x, y):
        """
        Set the semi-major-axis coordinate vector for inverse operations.

        Args:
            x, y (float) : Single values for use in calculating the
                on-sky focal plane coordinates.
        """
        self.D = numpy.array([ -self.xc, -self.yc, x, (1-self.ell)*y ])


    def _calculate_polar(self, x, y):
        r"""
        Calculate the polar coordinates (radius and azimuth) provided
        the Cartesian semi-major-axis coordinates :math:`(x_a,y_a)`
        using
        
        .. math::

            R &= \sqrt{x_a^2 + y_a^2} \\
            \theta &= \tan^{-1}\left(\frac{-y_a}{x_a}\right)

        Args:
            x,y (array-like): The semi-major-axis Cartesian coordinates
                :math:`(x_a,y_a)`.

        Returns:
            numpy.ndarray: The semi-major-axis polar coordinates:
            :math:`R, \theta`.
        """
#        _x = numpy.asarray(x)
#        _y = numpy.asarray(y)
        _x = numpy.atleast_1d(x)
        _y = numpy.atleast_1d(y)
        if _x.size != _y.size:
            raise ValueError('X and Y arrays must have the same size')
        R = numpy.sqrt( _x*_x + _y*_y)
        theta = numpy.degrees( numpy.arctan2(-_y, _x) )
        if hasattr(theta, '__len__'):
            theta[theta < 0] += 360.
        elif theta < 0:
            theta += 360.
        # Returned range in theta is -pi,pi: convert to 0,2pi
        return R, theta
    

    def _calculate_cartesian(self, r, theta):
        r"""
        Invert the calculation of the semi-major-axis polar coordinates
        to calculate the semi-major-axis Cartesian coordinates
        :math:`(x_a,y_a)` using
        
        .. math::

            x_a &= \pm R / \sqrt{1 + \tan^2\theta}\\
            y_a &= -x_a\ \tan\theta

        where :math:`x_a` is negative when :math:`\pi/2 \leq \theta <
        3\pi/2`.

        Args:

            r,theta (array-like): The semi-major-axis polar coordinates
                :math:`(R,\theta)`.

        Returns:
            numpy.ndarray: The semi-major-axis Cartesian coordinates:
            :math:`x_a, y_a`.
        """
#        _r = numpy.asarray(r)
#        _theta = numpy.asarray(theta)
        _r = numpy.atleast_1d(r)
        _theta = numpy.atleast_1d(theta)
        if _r.size != _theta.size:
            raise ValueError('R and THETA arrays must have the same size')
        tant = numpy.tan(numpy.radians(_theta))
        xd = _r/numpy.sqrt(1.0 + numpy.square(tant))
        if hasattr(xd, '__len__'):
            xd[(_theta > 90) & (_theta <= 270)] *= -1.
        elif _theta > 90 and _theta <= 270:
            xd *= -1.
        yd = -xd*tant
        return xd, yd
    

    def solve(self, x, y):
        r"""
        Use `scipy.linalg.lu_solve`_ to solve :math:`{\mathbf x} =
        {\mathbf A}^{-1}\ {\mathbf b}`.

        Args:
            x,y (array-like): The coordinates :math:`(x_f,y_f)`, which
                are the sky-right, focal-plane Cartesian coordinates
                relative to a reference on-sky position
                :math:`(x_0,y_0)`, which is relative to the center of
                the ellipse (galaxy center).

        Returns:
            numpy.ndarray: The :math:`{\mathbf x}` vectors (separated by
            rows) as defined by the solution to :math:`{\mathbf A}^{-1}\
            {\mathbf b}`

        Raises:
            ValueError: Raised if object was not properly defined or if
                the X and Y arrays do not have the same size.
        """
        if not self._defined():
            raise ValueError('SemiMajorAxisCoo object not fully defined!')
#        _x = numpy.asarray(x)
#        _y = numpy.asarray(y)
        _x = numpy.atleast_1d(x)
        _y = numpy.atleast_1d(y)
        if _x.size != _y.size:
            raise ValueError('X and Y arrays must have the same size')
        out = numpy.empty((_x.size, 6), dtype=float)
        for i, (xx, yy) in enumerate(zip(_x.ravel(),_y.ravel())):
            self._setB(xx,yy)
            out[i,:] = linalg.lu_solve((self.Alu, self.Apiv), self.B)
        return out.reshape(_x.shape + (6,))


    def solve_inverse(self, x, y):
        r"""
        Use `scipy.linalg.lu_solve`_ to solve :math:`{\mathbf f} =
        {\mathbf C}^{-1}\ {\mathbf d}`.

        Args:
            x,y (array-like): The semi-major-axis Cartesian coordinates
                :math:`(x_a,y_a)`.

        Returns:
            numpy.ndarray: The :math:`{\mathbf f}` vector as defined by
            the solution to :math:`{\mathbf C}^{-1}\ {\mathbf d}`

        Raises:
            ValueError: Raised if object was not properly defined or if
                the X and Y arrays do not have the same size.
        """
        if not self._defined():
            raise ValueError('SemiMajorAxisCoo object not fully defined!')
#        _x = numpy.asarray(x)
#        _y = numpy.asarray(y)
        _x = numpy.atleast_1d(x)
        _y = numpy.atleast_1d(y)
        if _x.size != _y.size:
            raise ValueError('X and Y arrays must have the same size')
        out = numpy.empty((_x.size, 4), dtype=float)
        for i, (xx, yy) in enumerate(zip(_x.ravel(),_y.ravel())):
            self._setD(xx,yy)
            out[i,:] = linalg.lu_solve((self.Clu, self.Cpiv), self.D)
        return out.reshape(_x.shape + (4,))


    def coo(self, x, y):
        r"""
        Calculate :math:`{\mathbf x}` using :func:`solve` for the
        provided :math:`(x_f,y_f)` and return the semi-major-axis
        Cartesian and polar coordinates, :math:`(x_a,y_a)` and
        :math:`(R,\theta)`.  This combines the functionality of
        :func:`cartesian` and :func:`polar`, and so is more efficient
        than using these both separately.

        Args:
            x,y (array-like): The coordinates :math:`(x_f,y_f)`, which
                are the sky-right, focal-plane position relative to a
                reference on-sky position :math:`(x_0,y_0)` relative to
                the center of the ellipse (galaxy center),

        Returns:
            numpy.ndarray: Four arrays with the semi-major-axis
            Cartesian and polar coordinates:  :math:`x_a, y_a, R,
            \theta`.

        """
        coo = self.solve(x, y)
        return (coo[...,4], coo[...,5]) + self._calculate_polar(coo[...,4], coo[...,5])


    def polar(self, x, y):
        r"""
        Calculate :math:`{\mathbf x}` using :func:`solve` for the
        provided :math:`(x_f,y_f)` and return the semi-major-axis polar
        coordinates, :math:`(R,\theta)`, where
        
        .. math::

            R &= \sqrt{x_a^2 + y_a^2} \\
            \theta &= \tan^{-1}\left(\frac{-y_a}{x_a}\right)

        Args:
            x,y (array-like): The coordinate :math:`(x_f,y_f)`, which
                is the sky-right, focal-plane position relative to a
                reference on-sky position :math:`(x_0,y_0)` relative to
                the center of the ellipse (galaxy center),

        Returns:
            numpy.ndarray: Two arrays with the semi-major-axis polar
            coordinates: :math:`R, \theta`.
        """
        coo = self.solve(x, y)
        return self._calculate_polar(coo[...,4], coo[...,5])

    
    def polar_invert(self, r, theta):
        r"""
        Calculate :math:`{\mathbf f}` using :func:`solve` for the
        provided :math:`(R,\theta)` and return focal-plane cartesian
        coordinates :math:`(x_f,y_f)`.

        Args:
            r,theta (array-like): The semi-major-axis polar coordinates
                :math:`(R,\theta)`.

        Returns:
            numpy.ndarray: Two arrays with the focal-plane Cartesian
            coordinates :math:`(x_f,y_f)`.
        """
        xd, yd = self._calculate_cartesian(r, theta)
        coo = self.solve_inverse(xd, yd)
        return coo[...,0], coo[...,1]


    def cartesian(self, x, y):
        r"""
        Calculate :math:`{\mathbf x}` using :func:`solve` for the
        provided :math:`(x_f,y_f)` and return the semi-major-axis
        Cartesian and coordinates, :math:`(x_a,y_a)`.

        Args:
            x,y (array-like): The coordinate :math:`(x_f,y_f)`, which
                is the sky-right, focal-plane position relative to a
                reference on-sky position :math:`(x_0,y_0)` relative to
                the center of the ellipse (galaxy center),

        Returns:
            numpy.ndarray: Two arrays with the semi-major-axis Cartesian
            coordinates, :math:`x_a, y_a`.
        """
        coo = self.solve(x, y)
        return coo[...,4], coo[...,5]


    def cartesian_invert(self, x, y):
        r"""
        Calculate :math:`{\mathbf f}` using :func:`solve` for the
        provided :math:`(x_a,y_a)` and return focal-plane cartesian
        coordinates :math:`(x_f,y_f)`.

        Args:
            x,y (array-like): The semi-major-axis Cartesian coordinates
                :math:`(x_a,y_a)`.

        Returns:
            numpy.ndarray: The focal-plane Cartesian coordinates
            :math:`(x_f,y_f)`.
        """
        coo = self.solve_inverse(x, y)
        return coo[...,0], coo[...,1]


    






#!/usr/bin/env python3

import os
import time
from IPython import embed
import numpy
from astropy.io import fits
import astropy.constants
from matplotlib import pyplot

from mangadap.util.drpfits import DRPFitsBitMask, DRPFits
from mangadap.dapfits import DAPCubeBitMask
from mangadap.config import defaults

#-----------------------------------------------------------------------------

def collate_data(ofile, tbl, dapver=None, daptype=None):

    # The number of pixels is HARDWIRED!
    nspec = len(tbl)
    npix = 4563

    drpbm = DRPFitsBitMask()
    cube_bm = DAPCubeBitMask()
    flux = numpy.zeros((nspec, npix), dtype=float)
    ivar = numpy.zeros((nspec, npix), dtype=float)
    mask = numpy.zeros((nspec, npix), dtype=drpbm.minimum_dtype())
    model = numpy.zeros((nspec, npix), dtype=float)
    model_mask = numpy.zeros((nspec, npix), dtype=cube_bm.minimum_dtype())
    emline = numpy.zeros((nspec, npix), dtype=float)
    stellar = numpy.zeros((nspec, npix), dtype=float)
    stellar_mask = numpy.zeros((nspec, npix), dtype=cube_bm.minimum_dtype())
    lsf = numpy.zeros((nspec, npix), dtype=float)
    redshift = numpy.zeros(nspec, dtype=float)
    filled = numpy.zeros(nspec, dtype=bool)

    for s in range(nspec):
        print('{0}/{1}'.format(s+1, nspec), end='\r')
        if tbl['NBIN'][s] == 0:
            mask[s,:] = drpbm.turn_on(mask[s,:], 'DONOTUSE')
            filled[s] = True
            continue
        if filled[s]:
            continue

        # Open both the DRP datacube and the DAP model datacube
        drp_hdu = fits.open(os.path.join(*DRPFits.default_paths(tbl['PLT'][s], tbl['IFU'][s],
                                                                'CUBE')))
        dap_hdu = fits.open(os.path.join(
                            defaults.dap_method_path(daptype, plate=tbl['PLT'][s],
                                                     ifudesign=tbl['IFU'][s], dapver=dapver),
                            defaults.dap_file_name(tbl['PLT'][s], tbl['IFU'][s], daptype,
                                                   mode='LOGCUBE')))

        # Find any selected spectra in this datacube
        indx = numpy.where((tbl['PLT'] == tbl['PLT'][s]) & (tbl['IFU'] == tbl['IFU'][s]))[0]
        if len(indx) == 0:
            embed(header='could not find pltifu')
            exit()

        try:
            i, j = numpy.array([list(map(lambda x : x[0],
                                     numpy.where(dap_hdu['BINID'].data[1] == tbl['BIN'][ii]))) 
                            for ii in indx]).T
        except:
            embed(header='bad bin selection')
            exit()

        wave = drp_hdu['WAVE'].data
        flux[indx,:] = drp_hdu['FLUX'].data[:,i,j].T
        ivar[indx,:] = drp_hdu['IVAR'].data[:,i,j].T
        mask[indx,:] = drp_hdu['MASK'].data[:,i,j].T
        model[indx,:] = dap_hdu['MODEL'].data[:,i,j].T
        model_mask[indx,:] = dap_hdu['MODEL_MASK'].data[:,i,j].T
        emline[indx,:] = dap_hdu['EMLINE'].data[:,i,j].T
        stellar[indx,:] = dap_hdu['STELLAR'].data[:,i,j].T
        stellar_mask[indx,:] = dap_hdu['STELLAR_MASK'].data[:,i,j].T
        lsf[indx,:] = dap_hdu['LSF'].data[:,i,j].T
        redshift[indx] = tbl['Z'][indx]
        filled[indx] = True

    print('{0}/{0}'.format(nspec))
    print('All filled: {0}'.format(str(numpy.all(filled))))
    
    fits.HDUList([fits.PrimaryHDU(),
                  fits.ImageHDU(data=wave, name='WAVE'),
                  fits.ImageHDU(data=flux, name='FLUX'),
                  fits.ImageHDU(data=ivar, name='IVAR'),
                  fits.ImageHDU(data=mask, name='MASK'),
                  fits.ImageHDU(data=model, name='MODEL'),
                  fits.ImageHDU(data=model_mask, name='MODEL_MASK'),
                  fits.ImageHDU(data=emline, name='EMLINE'),
                  fits.ImageHDU(data=stellar, name='STELLAR'),
                  fits.ImageHDU(data=stellar_mask, name='STELLAR_MASK'),
                  fits.ImageHDU(data=lsf, name='LSF'),
                  fits.ImageHDU(data=redshift, name='Z')]).writeto(ofile, overwrite=True)


#-----------------------------------------------------------------------------

def main():
    ifile = 'representative_spectra_selection_v2.fits'
    ofile = 'benchmark_spectra_v2.fits'
    dapver = '3.0.1'
    daptype = 'SPX-MILESHC-MASTARHC2'

    # Construct a single file with all the "representative" spectra
    with fits.open(ifile) as hdu:
        collate_data(ofile, hdu['PAR'].data, dapver=dapver, daptype=daptype)
    
if __name__ == '__main__':
    t = time.perf_counter()
    main()
    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))







from pathlib import Path
import time

from IPython import embed

import numpy

from matplotlib import pyplot, ticker, rc, colors

from astropy.io import fits
import astropy.constants

from mangadap.util.bitmask import BitMask
from mangadap.dapfits import DAPQualityBitMask
from mangadap.config import defaults, manga
from mangadap.proc.util import growth_lim
from mangadap.config.analysisplan import AnalysisPlan
from mangadap.util.fileio import channel_dictionary

from mangadap.scripts import scriptbase

#-----------------------------------------------------------------------------
def init_ax(fig, pos, facecolor=None, grid=False):
    ax = fig.add_axes(pos, facecolor=facecolor)
    ax.minorticks_on()
    ax.tick_params(which='major', length=6) #, direction='in')
    ax.tick_params(which='minor', length=3) #, direction='in')
    if grid:
        ax.grid(True, which='major', color='0.8', zorder=1, linestyle='-', lw=0.5)
    return ax


def dapall_qa(drpall, dapall, analysis_path, daptype, eml, spi):
    # TODO: HARDCODED!
    oroot = analysis_path / daptype / 'qa'
    if not oroot.exists():
        oroot.mkdir(parents=True)

    # Plot H-delta A equivalent width versus D4000
    ew_d4000(drpall, dapall, daptype, eml, spi, ofile=str(oroot / 'dapall_ew_d4000.png'))

    # Plot the star-formation, mass diagram
    mass_lha(drpall, dapall, daptype, eml, ofile=str(oroot / 'dapall_mass_lha.png'))

    # Plot the approximate Faber-Jackson relation
    mass_sigma(drpall, dapall, daptype, ofile=str(oroot / 'dapall_mass_sigma.png'))

    # Plot the approximate Tully-Fisher relation
    velocity_gradient(drpall, dapall, daptype, eml, ofile=str(oroot / 'dapall_mass_vel.png'))

    # Plot the MgFe and Hbeta absorption indices
    mgfe_hbeta(drpall, dapall, daptype, spi, ofile=str(oroot / 'dapall_mgfe_hbeta.png'))

    # Plot the radial coverage of each cube
    radial_coverage_histogram(dapall, daptype, ofile=str(oroot / 'dapall_radialcoverage.png'))

    # Plot the redshift differences
    redshift_distribution(dapall, daptype, ofile=str(oroot / 'dapall_redshift_dist.png'))


def ew_d4000(drpall, dapall, daptype, eml, spi, ofile=None):

    indx = (dapall['DAPDONE'] == 1) & (dapall['DAPTYPE'] == daptype)
    indx &= (drpall['nsa_sersic_mass'][dapall['DRPALLINDX']] > 0)

    ha_indx = indx & (dapall['EMLINE_GEW_1RE'][:,eml['Ha-6564']] > -999) \
                    & (dapall['SPECINDEX_1RE'][:,spi['Dn4000']] > -999)
    hd_indx = indx & (dapall['SPECINDEX_1RE'][:,spi['HDeltaA']] > -999) \
                    & (dapall['SPECINDEX_1RE'][:,spi['Dn4000']] > -999)

    is_critical = DAPQualityBitMask().flagged(dapall['DAPQUAL'], 'CRITICAL')

    ha_bd = ha_indx & is_critical
    hd_bd = hd_indx & is_critical

    ha_gd = ha_indx & numpy.invert(is_critical)
    hd_gd = hd_indx & numpy.invert(is_critical)

    mass_lim = [1e9, 7e11]
    d4_lim = [0.9, 2.3]
    ha_lim = [2e-2, 400]
    hd_lim = [-4, 9]

    font = { 'size' : 12 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    ax = init_ax(fig, [0.3, 0.56, 0.4, 0.4])
    ax.set_xlim(d4_lim)
    ax.set_ylim(ha_lim)
    ax.tick_params(axis='x', which='both', direction='in')
    ax.set_yscale('log')
    ax.xaxis.set_major_formatter(ticker.NullFormatter())

    sc = ax.scatter(dapall['SPECINDEX_1RE'][ha_gd,spi['Dn4000']],
                    dapall['EMLINE_GEW_1RE'][ha_gd,eml['Ha-6564']],
                    lw=0, marker='.', s=20, zorder=3,
                    c=drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][ha_gd]],
                    norm=colors.LogNorm(vmin=mass_lim[0], vmax=mass_lim[1]),
                    cmap='viridis')

    ax.scatter(dapall['SPECINDEX_1RE'][ha_bd,spi['Dn4000']],
               dapall['EMLINE_GEW_1RE'][ha_bd,eml['Ha-6564']],
               lw=0, marker='x', s=10, zorder=2,
               c=drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][ha_bd]],
               norm=colors.LogNorm(vmin=mass_lim[0], vmax=mass_lim[1]),
               cmap='viridis')

    ax.text(-0.25, 0.5, r'$\langle {\rm EW}_{{\rm H}\alpha}\rangle_e$ [$\AA$, in emission]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')

    ax = init_ax(fig, [0.3, 0.15, 0.4, 0.4])
    ax.set_xlim(d4_lim)
    ax.set_ylim(hd_lim)

    sc = ax.scatter(dapall['SPECINDEX_1RE'][hd_gd,spi['Dn4000']],
                    dapall['SPECINDEX_1RE'][hd_gd,spi['HDeltaA']],
                    lw=0, marker='.', s=20, zorder=3,
                    c=drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][hd_gd]],
                    norm=colors.LogNorm(vmin=mass_lim[0], vmax=mass_lim[1]),
                    cmap='viridis')

    ax.scatter(dapall['SPECINDEX_1RE'][hd_bd,spi['Dn4000']],
               dapall['SPECINDEX_1RE'][hd_bd,spi['HDeltaA']],
               lw=0, marker='x', s=10, zorder=2,
               c=drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][hd_bd]],
               norm=colors.LogNorm(vmin=mass_lim[0], vmax=mass_lim[1]),
               cmap='viridis')

    cax = fig.add_axes([0.51, 0.48, 0.15, 0.01])
    cb = pyplot.colorbar(sc, cax=cax, orientation='horizontal', ticks=[1e9,1e10,1e11])
                         #, ticks=[0.1,1], format=ticker.ScalarFormatter())
    cax.text(0.5, 2.8, r'$\mathcal{M}_\ast$ [$h^{-2} \mathcal{M}_\odot$]',
             ha='center', va='center', transform=cax.transAxes) #, fontsize=10)

    ax.text(-0.25, 0.5, r'$\langle {\rm H}\delta_{\rm A}\rangle_e$ [$\AA$, in absorption]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')

    ax.text(0.5, -0.18, r'$\langle {\rm Dn4000}\rangle_e$',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)

    if ofile is None:
        pyplot.show()
    else:
        print('Writing: {0}'.format(ofile))
        fig.canvas.print_figure(ofile, bbox_inches='tight', dpi=100)
    fig.clear()
    pyplot.close(fig)


def mass_lha(drpall, dapall, daptype, eml, ofile=None):

    indx = (dapall['DAPDONE'] == 1) & (dapall['DAPTYPE'] == daptype)
    indx &= (drpall['nsa_sersic_mass'][dapall['DRPALLINDX']] > 0)

    indx &= (dapall['EMLINE_GSB_1RE'][:,eml['Hb-4862']] > -999)
    indx &= (dapall['EMLINE_GSB_1RE'][:,eml['Ha-6564']] > -999)
    indx &= (dapall['SFR_1RE'] > -999)

    decrement = numpy.ma.divide(dapall['EMLINE_GSB_1RE'][:,eml['Ha-6564']],
                                   dapall['EMLINE_GSB_1RE'][:,eml['Hb-4862']])

    loew_indx =  indx & (dapall['EMLINE_GEW_1RE'][:,eml['Ha-6564']] > -999)\
                        & (dapall['EMLINE_GEW_1RE'][:,eml['Ha-6564']] < 2)
    hiew_indx = indx & (dapall['EMLINE_GEW_1RE'][:,eml['Ha-6564']] > 2)

    is_critical = DAPQualityBitMask().flagged(dapall['DAPQUAL'], 'CRITICAL')

    loew_bd =  loew_indx & is_critical
    hiew_bd =  hiew_indx & is_critical

    loew_gd =  loew_indx & numpy.invert(is_critical)
    hiew_gd =  hiew_indx & numpy.invert(is_critical)


    mass_lim = [1e8, 1e12]
    lha_lim = [36.5, 42.5]

    font = { 'size' : 16 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    ax = init_ax(fig, [0.2, 0.2, 0.6, 0.6])
    ax.set_xlim(mass_lim)
    ax.set_ylim(lha_lim)
    ax.set_xscale('log')

    ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][loew_gd]],
               numpy.ma.log10(dapall['SFR_1RE'][loew_gd])+41.27,
               color='0.5', marker='.', s=20, lw=0, zorder=3, alpha=0.3)

    sc = ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][hiew_gd]],
                    numpy.ma.log10(dapall['SFR_1RE'][hiew_gd])+41.27,
                    c=decrement[hiew_gd], vmin=2.8, vmax=7,
                    marker='.', s=20, lw=0, zorder=4, cmap='inferno')

    ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][loew_bd]],
               numpy.ma.log10(dapall['SFR_1RE'][loew_bd])+41.27,
               color='0.5', marker='x', s=10, lw=0, zorder=3, alpha=0.3)

    ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][hiew_bd]],
               numpy.ma.log10(dapall['SFR_1RE'][hiew_bd])+41.27,
               c=decrement[hiew_bd], vmin=2.8, vmax=7,
               marker='x', s=10, lw=0, zorder=4, cmap='inferno')

    ax.text(-0.15, 0.5, r'$\log\ \langle L_{{\rm H}\alpha}\rangle_e$ [erg/s]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')

    ax.text(0.5, -0.12, r'$\mathcal{M}_\ast$ [$h^{-2} \mathcal{M}_\odot$]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)

    cax = fig.add_axes([0.58, 0.25, 0.20, 0.01])
    cb = pyplot.colorbar(sc, cax=cax, orientation='horizontal', ticks=[3, 4, 5, 6, 7])
    cax.text(0.5, 3.0, r'$L_{{\rm H}\alpha}/L_{{\rm H}\beta}$',
             ha='center', va='center', transform=cax.transAxes) #, fontsize=10)

    if ofile is None:
        pyplot.show()
    else:
        print('Writing: {0}'.format(ofile))
        fig.canvas.print_figure(ofile, bbox_inches='tight', dpi=100)
    fig.clear()
    pyplot.close(fig)


def mass_sigma(drpall, dapall, daptype, ofile=None):

    indx = (dapall['DAPDONE'] == 1) & (dapall['DAPTYPE'] == daptype)
    indx &= (dapall['STELLAR_SIGMA_1RE'] > 0)
    indx &= (dapall['NSA_ELPETRO_TH50_R'] > 0)
    indx &= (dapall['ADIST_Z'] > 0)

    is_critical = DAPQualityBitMask().flagged(dapall['DAPQUAL'], 'CRITICAL')
    bd =  indx & is_critical
    gd =  indx & numpy.invert(is_critical)

    rekpc = dapall['NSA_ELPETRO_TH50_R']*numpy.radians(1/3600)*1e3*dapall['ADIST_Z']

    font = { 'size' : 18 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    mass_lim = [7e7, 3e12]
    sigm_lim = [10,  800]

    inst = numpy.array([[mass_lim[0], 70],
                        [mass_lim[1], 70],
                        [mass_lim[1], sigm_lim[0]],
                        [mass_lim[0], sigm_lim[0]]])

    ax = init_ax(fig, [0.2, 0.20, 0.6, 0.6])
    ax.set_xlim(mass_lim)
    ax.set_ylim(sigm_lim)
    ax.set_xscale('log')
    ax.set_yscale('log')

    x = numpy.linspace(6, 14, 100)
    y = x/2 - 2.6
    ax.plot(numpy.power(10,x), numpy.power(10,y-2), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y-1), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y+1), color='0.8', zorder=0)

    ax.fill(inst[:,0], inst[:,1], color='0.1', alpha=0.1, zorder=1)

    sc = ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][gd]],
                    dapall['STELLAR_SIGMA_1RE'][gd], lw=0, marker='.', s=20, zorder=3,
                    c=rekpc[gd], norm=colors.LogNorm(vmin=1, vmax=30), cmap='inferno')

    ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][bd]],
               dapall['STELLAR_SIGMA_1RE'][bd], lw=0, marker='x', s=10, zorder=3,
               c=rekpc[bd], norm=colors.LogNorm(vmin=1, vmax=30), cmap='inferno')

    cax = fig.add_axes([0.63, 0.26, 0.15, 0.01])
    cb = pyplot.colorbar(sc, cax=cax, orientation='horizontal', ticks=[1,10],
                         format=ticker.ScalarFormatter())
    cax.text(0.5, 3.0, r'$R_e$ [kpc]', ha='center', va='center', transform=cax.transAxes)
    

    ax.text(0.5, -0.15, r'$\mathcal{M}_\ast$ [$h^{-2} \mathcal{M}_\odot$]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)
    ax.text(-0.2, 0.5, r'$\sigma_{\ast,e}$ [km/s]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')
    ax.text(0.51, 0.85, r'$\mathcal{M}_\ast\propto\ \sigma^2_{\ast,e}$',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation=53, zorder=4)
    ax.text(0.99, 0.37, r'$\sigma_{\ast,e}\ \lesssim\ \sigma_{\rm inst}$',
            horizontalalignment='right', verticalalignment='center', transform=ax.transAxes,
            zorder=4)

    if ofile is None:
        pyplot.show()
    else:
        print('Writing: {0}'.format(ofile))
        fig.canvas.print_figure(ofile, bbox_inches='tight', dpi=100)
    fig.clear()
    pyplot.close(fig)


def velocity_gradient(drpall, dapall, daptype, eml, ofile=None):

    indx = (dapall['DAPDONE'] == 1) & (dapall['DAPTYPE'] == daptype)
    indx &= (drpall['nsa_elpetro_ba'][dapall['DRPALLINDX']] > 0)

    ha_indx = indx & (dapall['HA_GVEL_LO_CLIP'] > -999) & (dapall['HA_GVEL_HI_CLIP'] > -999)

    ha_vdiff = dapall['HA_GVEL_HI_CLIP'] - dapall['HA_GVEL_LO_CLIP']
    ha_sini = numpy.ma.sqrt(1-numpy.square(drpall['nsa_elpetro_ba'][dapall['DRPALLINDX']]))

    st_indx = indx & (dapall['STELLAR_VEL_LO_CLIP'] > -999) & (dapall['STELLAR_VEL_HI_CLIP'] > -999)

    st_vdiff = dapall['STELLAR_VEL_HI_CLIP'] - dapall['STELLAR_VEL_LO_CLIP']
    st_sini = numpy.ma.sqrt(1-numpy.square(drpall['nsa_elpetro_ba'][dapall['DRPALLINDX']]))

    is_critical = DAPQualityBitMask().flagged(dapall['DAPQUAL'], 'CRITICAL')

    ha_bd =  ha_indx & is_critical
    st_bd =  st_indx & is_critical

    ha_gd =  ha_indx & numpy.invert(is_critical)
    st_gd =  st_indx & numpy.invert(is_critical)

    mass_lim = [2e8, 5e11]
    velw_lim = [30,  4000]

    ha_sb_lim = [0.02, 8]
    st_sb_lim = [0.03, 1]

    font = { 'size' : 14 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    ax = init_ax(fig, [0.3, 0.56, 0.4, 0.4])
    ax.set_xlim(mass_lim)
    ax.set_ylim(velw_lim)
    ax.tick_params(axis='x', which='both', direction='in')
    ax.set_xscale('log')
    ax.set_yscale('log')
    ax.xaxis.set_major_formatter(ticker.NullFormatter())

    x = numpy.linspace(6, 14, 100)
    y = x/4
    ax.plot(numpy.power(10,x), numpy.power(10,y-0.75), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y+0.75), color='0.8', zorder=0)

    sc = ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][ha_gd]],
               numpy.ma.divide(ha_vdiff[ha_gd], ha_sini[ha_gd]),
               c=dapall['EMLINE_GSB_1RE'][ha_gd,eml['Ha-6564']],
               lw=0, marker='.', s=20, zorder=2,
               norm=colors.LogNorm(vmin=ha_sb_lim[0], vmax=ha_sb_lim[1]),
               cmap='inferno_r')
    ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][ha_bd]],
               numpy.ma.divide(ha_vdiff[ha_bd], ha_sini[ha_bd]),
               c=dapall['EMLINE_GSB_1RE'][ha_bd,eml['Ha-6564']],
               lw=0, marker='x', s=10, zorder=2,
               norm=colors.LogNorm(vmin=ha_sb_lim[0], vmax=ha_sb_lim[1]),
               cmap='inferno_r')

    cax = fig.add_axes([0.54, 0.60, 0.15, 0.01])
    cb = pyplot.colorbar(sc, cax=cax, orientation='horizontal', ticks=[0.1,1],
                         format=ticker.ScalarFormatter())
    pyplot.setp(cb.ax.xaxis.get_ticklabels(), fontsize=10)
    cax.text(0.8, 2.5, r'$\langle I_{{\rm H}\alpha}\rangle_e$', ha='center', va='center',
             transform=cax.transAxes, fontsize=10)

    ax.text(-0.22, 0.5, r'$\Delta V_{\rm gas}$ [km/s]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')


    ax = init_ax(fig, [0.3, 0.15, 0.4, 0.4])
    ax.set_xlim(mass_lim)
    ax.set_ylim(velw_lim)
    ax.set_xscale('log')
    ax.set_yscale('log')

    x = numpy.linspace(6, 14, 100)
    y = x/4
    ax.plot(numpy.power(10,x), numpy.power(10,y-0.75), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y), color='0.8', zorder=0)
    ax.plot(numpy.power(10,x), numpy.power(10,y+0.75), color='0.8', zorder=0)

    sc = ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][st_gd]],
                    numpy.ma.divide(st_vdiff[st_gd], st_sini[st_gd]), c=dapall['SB_1RE'][st_gd],
                    lw=0, marker='.', s=20, zorder=2,
                    norm=colors.LogNorm(vmin=st_sb_lim[0], vmax=st_sb_lim[1]),
                    cmap='inferno_r')
    ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][st_bd]],
               numpy.ma.divide(st_vdiff[st_bd], st_sini[st_bd]), c=dapall['SB_1RE'][st_bd],
               lw=0, marker='x', s=10, zorder=2,
               norm=colors.LogNorm(vmin=st_sb_lim[0], vmax=st_sb_lim[1]),
               cmap='inferno_r')

    cax = fig.add_axes([0.31, 0.50, 0.11, 0.01])
    cb = pyplot.colorbar(sc, cax=cax, orientation='horizontal', ticks=[0.1,1],
                         format=ticker.ScalarFormatter())
    pyplot.setp(cb.ax.xaxis.get_ticklabels(), fontsize=10)
    cax.text(0.5, 2.5, r'$\langle I_g\rangle_e$', ha='center', va='center',
             transform=cax.transAxes, fontsize=10)

    ax.text(0.5, -0.18, r'$\mathcal{M}_\ast$ [$h^{-2} \mathcal{M}_\odot$]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)
    ax.text(-0.22, 0.5, r'$\Delta V_\ast$ [km/s]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')
    ax.text(0.55, 0.89, r'$\mathcal{M}_\ast\propto\ \Delta V^4$',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation=20, zorder=4)

    if ofile is None:
        pyplot.show()
    else:
        print('Writing: {0}'.format(ofile))
        fig.canvas.print_figure(ofile, bbox_inches='tight', dpi=100)
    fig.clear()
    pyplot.close(fig)


def mgfe_hbeta(drpall, dapall, daptype, spi, ofile=None):
    indx = (dapall['DAPDONE'] == 1) & (dapall['DAPTYPE'] == daptype)
    indx &= (drpall['nsa_sersic_mass'][dapall['DRPALLINDX']] > 0)

    indx &= (dapall['SPECINDEX_1RE'][:,spi['Mgb']] > -999)
    indx &= (dapall['SPECINDEX_1RE'][:,spi['Fe5270']] > -999)
    indx &= (dapall['SPECINDEX_1RE'][:,spi['Fe5335']] > -999)
    indx &= (dapall['SPECINDEX_1RE'][:,spi['Hb']] > -999)

    is_critical = DAPQualityBitMask().flagged(dapall['DAPQUAL'], 'CRITICAL')
    bd =  indx & is_critical
    gd =  indx & numpy.invert(is_critical)

    mgfe = numpy.ma.sqrt(dapall['SPECINDEX_1RE'][:,spi['Mgb']]
                            * (0.72*dapall['SPECINDEX_1RE'][:,spi['Fe5270']] 
                                + 0.28*dapall['SPECINDEX_1RE'][:,spi['Fe5335']]))

    mass_lim = [1e9, 7e11]
    mgfe_lim = [0, 4.5]
    hb_lim = [0.5, 5.9]

    font = { 'size' : 16 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    ax = init_ax(fig, [0.2, 0.2, 0.6, 0.6])
    ax.set_xlim(mgfe_lim)
    ax.set_ylim(hb_lim)
    ax.tick_params(axis='x', which='both', direction='in')

    sc = ax.scatter(mgfe[gd], dapall['SPECINDEX_1RE'][gd,spi['Hb']],
                    marker='.', s=30, lw=0, zorder=2,
                    c=drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][gd]],
                    norm=colors.LogNorm(vmin=mass_lim[0], vmax=mass_lim[1]),
                    cmap='viridis')
    ax.scatter(mgfe[bd], dapall['SPECINDEX_1RE'][bd,spi['Hb']],
               marker='x', s=15, lw=0, zorder=2,
               c=drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][bd]],
               norm=colors.LogNorm(vmin=mass_lim[0], vmax=mass_lim[1]),
               cmap='viridis')

    ax.text(-0.15, 0.5, r'$\langle {\rm H}\beta\rangle_e$ [$\AA$, in absorption]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')

    ax.text(0.5, -0.12, r'[$\langle$Mgb$\rangle_e\cdot$(0.72 $\langle$Fe5270$\rangle_e$ + 0.28 $\langle$Fe5335$\rangle_e$)]$^{1/2}$',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)
    ax.text(0.5, -0.19, r'[$\AA$, in absorption]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)

    cax = fig.add_axes([0.58, 0.71, 0.20, 0.01])
    cb = pyplot.colorbar(sc, cax=cax, orientation='horizontal', ticks=[1e9,1e10,1e11])
    cax.text(0.5, 3.0, r'$\mathcal{M}_\ast$ [$h^{-2} \mathcal{M}_\odot$]',
             ha='center', va='center', transform=cax.transAxes) #, fontsize=10)

    if ofile is None:
        pyplot.show()
    else:
        print('Writing: {0}'.format(ofile))
        fig.canvas.print_figure(ofile, bbox_inches='tight', dpi=100)
    fig.clear()
    pyplot.close(fig)


def radial_coverage_histogram(dapall, daptype, ofile=None):

    # Must have finished, be of the correct daptype, and *not* be
    # critical
    indx = (dapall['DAPDONE'] == 1) & (dapall['DAPTYPE'] == daptype) \
                & numpy.invert(DAPQualityBitMask().flagged(dapall['DAPQUAL'], 'CRITICAL'))

    # Get the range
    arcsec_rng = growth_lim(dapall['RCOV90'][indx], 0.99, fac=1.1)
    rore = numpy.ma.divide(dapall['RCOV90'][indx], dapall['NSA_ELPETRO_TH50_R'][indx])
    rore[dapall['NSA_ELPETRO_TH50_R'][indx] < 0] = numpy.ma.masked
    rore_rng = growth_lim(rore, 0.99, fac=1.1)

    # Determine the sample
    sdssMaskbits = defaults.sdss_maskbits_file()
    targ1bm = BitMask.from_par_file(str(sdssMaskbits), 'MANGA_TARGET1')

    ancillary = dapall['MNGTARG3'][indx] > 0
    primaryplus = targ1bm.flagged(dapall['MNGTARG1'][indx],
                                  flag=['PRIMARY_v1_2_0', 'COLOR_ENHANCED_v1_2_0'])
    secondary = targ1bm.flagged(dapall['MNGTARG1'][indx], flag='SECONDARY_v1_2_0')
    other = numpy.invert(ancillary) & numpy.invert(primaryplus) & numpy.invert(secondary)

    print('Num. of observations with 90% coverage <1 arcsec: {0}'.format(
            numpy.sum(dapall['RCOV90'][indx] < 1)))

    font = { 'size' : 10 }
    rc('font', **font)

    # Instantiate the figure
    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    nlim = [0.8, 1000]

    # Show RCOV90 in terms of arcseconds for each IFU
    ax = init_ax(fig, [0.1, 0.58, 0.8, 0.4])
    ax.set_ylim(nlim)
    ax.set_yscale('log', nonpositive='clip')
    ifu = [ 19, 37, 61, 91, 127 ]
    for i,f in enumerate(ifu):
        _indx = indx & ((dapall['IFUDESIGN']/100).astype(int) == f)
        if not numpy.any(_indx):
            continue
        ax.hist(dapall['RCOV90'][_indx], bins=50, range=arcsec_rng, alpha=0.5,
                histtype='stepfilled', color='C{0}'.format(i), zorder=2)
        ax.hist(dapall['RCOV90'][_indx], bins=50, range=arcsec_rng,
                histtype='step', color='C{0}'.format(i), zorder=3)
        medp = numpy.median(dapall['RCOV90'][_indx])
        ax.plot([medp, medp], nlim, color='C{0}'.format(i), zorder=4, linestyle='--')
    ax.text(-0.1, 0.5, 'N', ha='center', va='center', transform=ax.transAxes, rotation='vertical')
    ax.text(0.5, -0.13, r'$R_{90}$ (arcsec)', ha='center', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.95, '19-fiber', color='C0', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.90, '37-fiber', color='C1', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.85, '61-fiber', color='C2', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.80, '91-fiber', color='C3', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.75, '127-fiber', color='C4', ha='left', va='center', transform=ax.transAxes)

    # Show RCOV90 in terms of Re for each sample
    ax = init_ax(fig, [0.1, 0.08, 0.8, 0.4])
    ax.set_ylim(nlim)
    ax.set_yscale('log', nonpositive='clip')

    if numpy.any(primaryplus):
        _rore = (rore[primaryplus]).compressed()
        ax.hist(_rore, bins=50, range=rore_rng, alpha=0.5, histtype='stepfilled',
                color='C0', zorder=2)
        ax.hist(_rore, bins=50, range=rore_rng, histtype='step',
                color='C0', zorder=3)
        medp = numpy.median(_rore)
        ax.plot([medp, medp], nlim, color='C0', zorder=4, linestyle='--')

    if numpy.any(secondary):
        _rore = (rore[secondary]).compressed()
        ax.hist(_rore, bins=50, range=rore_rng, alpha=0.5, histtype='stepfilled',
                color='C1', zorder=2)
        ax.hist(_rore, bins=50, range=rore_rng, histtype='step',
                color='C1', zorder=3)
        medp = numpy.median(_rore)
        ax.plot([medp, medp], nlim, color='C1', zorder=4, linestyle='--')

    if numpy.any(other):
        _rore = (rore[other]).compressed()
        ax.hist(_rore, bins=50, range=rore_rng, alpha=0.5, histtype='stepfilled',
                color='C2', zorder=2)
        ax.hist(_rore, bins=50, range=rore_rng, histtype='step',
                color='C2', zorder=3)

    if numpy.any(ancillary):
        _rore = (rore[ancillary]).compressed()
        ax.hist(_rore, bins=50, range=rore_rng, alpha=0.5, histtype='stepfilled',
                color='C3', zorder=2)
        ax.hist(_rore, bins=50, range=rore_rng, histtype='step',
                color='C3', zorder=3)

    ax.text(-0.1, 0.5, 'N', ha='center', va='center', transform=ax.transAxes, rotation='vertical')
    ax.text(0.5, -0.12, r'$R_{90}/R_{eff}$', ha='center', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.95, 'Primary+', color='C0', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.90, 'Secondary', color='C1', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.85, 'Other', color='C2', ha='left', va='center', transform=ax.transAxes)
    ax.text(0.05, 0.80, 'Ancillary', color='C3', ha='left', va='center', transform=ax.transAxes)

    if ofile is None:
        pyplot.show()
    else:
        print('Writing: {0}'.format(ofile))
        fig.canvas.print_figure(ofile, bbox_inches='tight')
    fig.clear()
    pyplot.close(fig)


def redshift_distribution(dapall, daptype, ofile=None):

    indx = (dapall['DAPDONE'] == 1) & (dapall['DAPTYPE'] == daptype)
    ns_indx = indx & (dapall['NSA_Z'] > -999) & (dapall['STELLAR_Z'] > -999)
    gs_indx = indx & (dapall['HA_Z'] > -999) & (dapall['STELLAR_Z'] > -999)

    is_critical = DAPQualityBitMask().flagged(dapall['DAPQUAL'], 'CRITICAL')

    ns_bd = ns_indx & is_critical
    gs_bd = gs_indx & is_critical

    ns_gd = ns_indx & numpy.invert(is_critical)
    gs_gd = gs_indx & numpy.invert(is_critical)


    font = { 'size' : 14 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    offset_lim = [-90, 90]
    ax = init_ax(fig, [0.17, 0.10, 0.56, 0.3])
    ax.set_xlim([-0.01, 0.16])
    ax.set_ylim(offset_lim)
    ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')

    dcz = astropy.constants.c.to('km/s').value * (dapall['STELLAR_Z'] - dapall['NSA_Z'])

    ax.scatter(dapall['NSA_Z'][ns_gd], dcz[ns_gd], marker='.', s=20, lw=0, color='k', zorder=3)
    ax.scatter(dapall['NSA_Z'][ns_bd], dcz[ns_bd], marker='.', s=20, lw=0, color='C3', zorder=3)
    ax.text(0.5, -0.2, r'$z_{\rm NSA}$', horizontalalignment='center',
             verticalalignment='center', transform=ax.transAxes)
    ax.text(-0.18, 0.5, r'$c (z_\ast - z_{\rm NSA})$ [km/s]', horizontalalignment='center',
             verticalalignment='center', transform=ax.transAxes, rotation='vertical')

    ax = init_ax(fig, [0.73, 0.10, 0.1, 0.3])
    ax.set_ylim(offset_lim)
    ax.tick_params(which='both', direction='in', bottom=False)
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.grid(True, which='major', axis='y', color='0.8', zorder=0, linestyle='-')
    ax.hist(dcz[ns_gd], range=offset_lim, bins=100, alpha=0.7, color='0.4', histtype='stepfilled',
            orientation='horizontal', zorder=3)
    ax.hist(dcz[ns_gd], range=offset_lim, bins=100, alpha=0.7, color='k', histtype='step', lw=0.5,
            orientation='horizontal', zorder=4)

    ax = init_ax(fig, [0.22, 0.50, 0.56, 0.3])
    ax.set_xlim(offset_lim)
    ax.tick_params(which='both', left=False)
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    ax.grid(True, which='major', axis='x', color='0.8', zorder=0, linestyle='-')

    dcz = astropy.constants.c.to('km/s').value * (dapall['STELLAR_Z'] - dapall['HA_Z'])

    ax.hist(dcz[gs_gd], range=offset_lim, bins=100, alpha=0.7, color='0.4', histtype='stepfilled',
            zorder=3)
    ax.hist(dcz[gs_gd], range=offset_lim, bins=100, alpha=0.7, color='k', histtype='step', lw=0.5,
            zorder=4)
    ax.text(0.5, -0.2, r'$c (z_\ast - z_{\rm gas})$ [km/s]', horizontalalignment='center',
             verticalalignment='center', transform=ax.transAxes)

    if ofile is None:
        pyplot.show()
    else:
        print('Writing: {0}'.format(ofile))
        fig.canvas.print_figure(ofile, bbox_inches='tight')
    fig.clear()
    pyplot.close(fig)


class DapAllQA(scriptbase.ScriptBase):

    @classmethod
    def name(cls):
        """
        Return the name of the executable.
        """
        return 'dapall_qa'

    @classmethod
    def get_parser(cls, width=None):

        parser = super().get_parser(description='Construct QA plots using the DAPall metadata',
                                    width=width)

        parser.add_argument('--drpver', type=str, help='DRP version', default=None)
        parser.add_argument('--dapver', type=str, help='DAP version', default=None)
        parser.add_argument("--redux_path", type=str, help="main DRP output path", default=None)
        parser.add_argument("--analysis_path", type=str, help="main DAP output path", default=None)

        parser.add_argument("--plan_file", type=str, help="parameter file with the MaNGA DAP "
                            "execution plan to use instead of the default" , default=None)

#        parser.add_argument('--daptype', type=str, help='DAP processing type', default=None)
        parser.add_argument('--normal_backend', dest='bgagg', action='store_false', default=True)

        return parser

    @staticmethod
    def main(args):
        t = time.perf_counter()

        if args.bgagg:
            pyplot.switch_backend('agg')

        redux_path = manga.drp_redux_path(drpver=args.drpver) \
                            if args.redux_path is None else Path(args.redux_path).resolve()
        analysis_path = manga.dap_analysis_path(drpver=args.drpver, dapver=args.dapver) \
                            if args.analysis_path is None else Path(args.analysis_path).resolve()

        plan = manga.MaNGAAnalysisPlan.default(analysis_path=analysis_path) \
                    if args.plan_file is None \
                    else manga.MaNGAAnalysisPlan.from_toml(args.plan_file,
                                                           analysis_path=analysis_path)

        drpall_file = manga.drpall_file(drpver=args.drpver, redux_path=redux_path)
        dapall_file = manga.dapall_file(drpver=args.drpver, dapver=args.dapver,
                                           analysis_path=analysis_path)

        drpall_hdu = fits.open(str(drpall_file))
        dapall_hdu = fits.open(str(dapall_file))

        eml = channel_dictionary(dapall_hdu, 0, prefix='ELG')
        spi = channel_dictionary(dapall_hdu, 0, prefix='SPI')

        for i, key in enumerate(plan.keys()):
            dapall_qa(drpall_hdu['MANGA'].data, dapall_hdu[plan[key]['key']].data, analysis_path,
                      plan[key]['key'], eml, spi)

        print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))

